#!/usr/bin/env python3
import pycurl
import re
import os
import os.path
import subprocess
import glob
import time
import sys

# -------------------------------------------------------------------
# Start editing here.
# -------------------------------------------------------------------
# Edit script settings here, please take care to  change things accordingly
# General settings:

# Script user-agent string (adapt to service..)
USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; rv:67.0) Gecko/20100101 Firefox/67.0"  # Adapt accordingly

# Download online stream (channel) list url
PLAYLIST_URL = "https://raw.githubusercontent.com/Free-IPTV/Countries/master/01_German.m3u"

# PLAYLIST_ENCODING = Set encondig for playlist selection and display
# PLAYLIST_ENCODING = 'utf-8'
PLAYLIST_ENCODING = 'iso-8859-14' # Enable for German Umlauts

# ffmepg directory/command bin or exe - use and link to "ffmepg.exe" or "./ffmpeg"
# Example linux
# FFMPEG_DIRECTORY = 'home/USERNAME/ffmpeg/./ffmpeg' OR "ffmpeg"
# Example windows
# FFMPEG_DIRECTORY = 'C:\\Users\\USERNAME\\Desktop\\ffmpeg-20190918-53d31e9-win64-static\\bin\\ffmpeg' # example windows
FFMPEG_DIRECTORY = 'ffmpeg'

# Set everything here to either "True" or "False", even if not using Windows or Linux
USE_LINUX_AT = True  # Use Linux "at" command? True/False, set to False if not using Linux AT


# Stuff below does not yet work!

USE_WINDOWS_TASK_SCHEDULER = False  # Use True/False

# If windows, on Linux, no need to change
# only used with "USE_WINDOWS_TASK_SCHEDULER = True"
# "7" = Windows 7, "8" = Windows 8, "10" = Win 10
WINDOWS_VERSION = "7"

# -------------------------------------------------------------------
# Stop editing here.
# -------------------------------------------------------------------


APP_FOLDER = os.getcwd()
TASK_XML = '<?xml version="1.0" encoding="UTF-16"?> \
<Task version="1.3" xmlns="http://schemas.microsoft.com/windows/2004/02/mit/task"> \
  <RegistrationInfo> \
    <Date>_REGISTRATION_DATE_</Date> \
    <Author>rec.py</Author> \
    <URI>\StreamDownload</URI> \
  </RegistrationInfo> \
  <Triggers> \
    <TimeTrigger> \
      <StartBoundary>_TIMESTART_</StartBoundary> \
      <EndBoundary>_TIMEEND_</EndBoundary> \
      <ExecutionTimeLimit>PT24H</ExecutionTimeLimit> \
      <Enabled>true</Enabled> \
    </TimeTrigger> \
  </Triggers> \
  <Principals> \
    <Principal id="Author"> \
      <UserId>S-1-5-21-2205445863-1784293702-2949643585-1001</UserId> \
      <LogonType>S4U</LogonType> \
      <RunLevel>LeastPrivilege</RunLevel> \
    </Principal> \
  </Principals> \
  <Settings> \
    <MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy> \
    <DisallowStartIfOnBatteries>false</DisallowStartIfOnBatteries> \
    <StopIfGoingOnBatteries>true</StopIfGoingOnBatteries> \
    <AllowHardTerminate>true</AllowHardTerminate> \
    <StartWhenAvailable>false</StartWhenAvailable> \
    <RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable> \
    <IdleSettings> \
      <StopOnIdleEnd>true</StopOnIdleEnd> \
      <RestartOnIdle>false</RestartOnIdle> \
    </IdleSettings> \
    <AllowStartOnDemand>true</AllowStartOnDemand> \
    <Enabled>true</Enabled> \
    <Hidden>false</Hidden> \
    <RunOnlyIfIdle>false</RunOnlyIfIdle> \
    <DisallowStartOnRemoteAppSession>false</DisallowStartOnRemoteAppSession> \
    <UseUnifiedSchedulingEngine>true</UseUnifiedSchedulingEngine> \
    <WakeToRun>true</WakeToRun> \
    <ExecutionTimeLimit>PT2H</ExecutionTimeLimit> \
    <DeleteExpiredTaskAfter>P1D</DeleteExpiredTaskAfter> \
    <Priority>7</Priority> \
  </Settings> \
  <Actions Context="Author"> \
    <Exec> \
      <Command>_COMMAND_</Command> \
      <Arguments>_COMMANDARGS_</Arguments> \
      <WorkingDirectory>' + APP_FOLDER + '</WorkingDirectory> \
    </Exec> \
  </Actions> \
</Task>'

# This is just a temporary storage for the downloaded playlist, but this file wont be removed by itself
PLAYLIST_FILE = "streamTempSave.m3u8"
PLAYLIST_DATA = ""
PLAYLIST_CHANNELS = {}
RECORDING_NAME = ""
RECORDING_CHANNEL_NAME = ""
RECORDING_CHANNEL_URL = ""
RECORDING_TIME = 0
RECORDING_DURATION = 0


def getXmlTimeString(timestamp):
    return time.strftime("%Y-%m-%dT%H:%M:00", time.localtime(timestamp))


def downloadPlaylist(downloadURL):
    global PLAYLIST_URL
    PLAYLIST_URL = downloadURL
    playlistFile = open(PLAYLIST_FILE, 'wb')

    print('Starting download of URI: \'%s\'\n' % downloadURL)
    c = pycurl.Curl()
    c.setopt(pycurl.URL, downloadURL)
    c.setopt(pycurl.USERAGENT, USER_AGENT)
    c.setopt(pycurl.VERBOSE, False)
    c.setopt(pycurl.ENCODING, PLAYLIST_ENCODING)
    c.setopt(pycurl.WRITEDATA, playlistFile)
    c.setopt(pycurl.NOPROGRESS, True)
    c.setopt(pycurl.TIMEOUT, 30)

    try:
        c.perform()
        print("\nSuccessfully downloaded playlist.\n")
        playlistFile.close()
        c.close()
        return True
    except Exception as error:
        playlistFile.close()
        c.close()
        print("Encountered an error:")
        raise error
    return False


def getStreamSelection():
    global PLAYLIST_CHANNELS, RECORDING_CHANNEL_NAME, RECORDING_CHANNEL_URL

    playlistFile = open(PLAYLIST_FILE, "r", encoding=PLAYLIST_ENCODING, errors='strict')
    data = playlistFile.readlines()

    currentChannel = None
    for line in data:
        line = line.rstrip()
        if line.startswith('#EXTINF:'):
            try:
                currentChannel = list(re.findall(r'tvg-name\=\"(http:|https:)(.[^"]*)', line))[0].strip()
            except:
                currentChannel = line.split(',', 2)[1].strip()

            currentChannel = bytes(
                currentChannel, PLAYLIST_ENCODING).decode(PLAYLIST_ENCODING, 'strict').rstrip()

            if line.find('radio="true"') != -1:
                currentChannel += ' (Radio)'

        elif (currentChannel != None and line.startswith('http')):
            if currentChannel.startswith('*'):
                continue

            PLAYLIST_CHANNELS[currentChannel] = bytes(
                line, PLAYLIST_ENCODING).decode(PLAYLIST_ENCODING, 'strict').rstrip()
            currentChannel = None

    sortedChannels = sorted(PLAYLIST_CHANNELS.keys())

    print("\nChannel selection:")
    for index, channel in enumerate(sortedChannels):
        print("%2d - %s" % (index, channel))

    print("x - Exit")

    print("\n")

    while True:
        channelNum = input("\nInput channel number for grabbing: ")
        if channelNum.lower() == "x":
            exit(0)
        else:
            try:
                RECORDING_CHANNEL_URL = PLAYLIST_CHANNELS[sortedChannels[int(
                    channelNum)]].rstrip()
                RECORDING_CHANNEL_NAME = sortedChannels[int(channelNum)]

                print("Selected recording channel: %s" %
                      (sortedChannels[int(channelNum)]))
                isValidSelection = input('Is the channel correct Y/N? ')
                if isValidSelection.lower() == "y":
                    return True
                else:
                    continue

            except Exception:
                print('Wrong channel selection. Enter a channel number or "x" to exit.')
                continue
            break


if __name__ == "__main__":
    hasPlaylist = False
    ISWEBDL = False
    if len(sys.argv) == 1:
        if downloadPlaylist(PLAYLIST_URL):
            hasPlaylist = True
            ISWEBDL = True
            print("Playlist succesfully downloaded.\n")
        else:
            print('Cannot download playlist "%s"' % (PLAYLIST_URL))
            exit(1)
    else:
        for opt in sys.argv:
            if opt.startswith('-local'):
                playlistName = opt.replace('-local', '', 1)
                if os.path.exists(playlistName):
                    print('Using local playlist: %s' % (playlistName))
                    PLAYLIST_FILE = playlistName
                    hasPlaylist = True
                    ISWEBDL = False
                else:
                    print('Playlist not found using name "%s": ' %(playlistName))
                    exit(1)
            elif opt.startswith('-pl'):
                global playlistURL
                playlistURL = opt.replace('-pl', '', 1)
                if downloadPlaylist(playlistURL):
                    hasPlaylist = True
                    ISWEBDL = True
                    print("Playlist succesfully downloaded.\n")
                else:
                    print('Cannot download playlist "%s"' %(PLAYLIST_URL))
                    exit(1)


    if hasPlaylist and getStreamSelection():
        print('\nIt\'s time to enter the date and time for the recording:')
        while True:
            while True:
                recordDate = input('\nEnter record date for recording, in format day.month.year, for example: 7.8.2019 or "x" to cancel, enter nothing for todays date.\n')
                if recordDate.lower() == 'x':
                    print("Cancelled")
                    exit(1)

                elif len(recordDate) == 0:
                    timeNow = time.gmtime()
                    timestampNow = time.strptime("%d.%d.%d 00:00:00" %(timeNow.tm_mday, timeNow.tm_mon, timeNow.tm_year), "%d.%m.%Y %H:%M:%S")
                    recordDay = timeNow.tm_mday
                    recordMonth = timeNow.tm_mon
                    recordYear = timeNow.tm_year

                    print("Record date: %02d.%02d.%d " %(timeNow.tm_mday, timeNow.tm_mon, timeNow.tm_year))
                    if input("Is the record date correct Y/N ? ").lower() != 'y':
                        continue
                    break

                elif recordDate.count('.') == 2:
                    recordDay, recordMonth, recordYear = map(int, recordDate.split('.', 3))
                    timeNow = time.gmtime()
                    timestampNow = time.strptime("%d.%d.%d 00:00:00" %(timeNow.tm_mday, timeNow.tm_mon, timeNow.tm_year), "%d.%m.%Y %H:%M:%S")

                    try:
                        timestampRecord = time.strptime("%d.%d.%d 00:00:00" %(recordDay, recordMonth, recordYear), "%d.%m.%Y %H:%M:%S")
                    except ValueError:
                        print('Date input does not match, use day.month.year as: 7.8.2019')
                        continue

                    if time.mktime(timestampNow) > time.mktime(timestampRecord):
                        print("Date is in the past. Repeat input.")
                    else:
                        print("Record date: %02d.%02d.%d " %(recordDay, recordMonth, recordYear))

                    if input("Is the record date correct Y/N ? ").lower() != 'y':
                        continue
                    break
                else:
                    print("Invalid date input.")
                    continue

            while True:
                recordTime = input('\nEnter time to start recording, in format hour:minute, for example: 14:00 for 14 o\'clock - limit: 23:59 (on the same day), "x" to cancel\n')
                if recordTime.lower() == 'x':
                    print("Cancelled")
                    exit(1)
                elif recordTime.find(':') != -1:
                    recordHour, recordMinute = map(int, recordTime.split(':', 2))
                    print("Record time: %02d:%02d " %(recordHour, recordMinute))

                    if (recordHour >= 24):
                        print("Record hour > 23 - limit is '23:59' for the same day, '0:0' is at 24 o'clock midnight, at the same day.")
                        continue
                    elif (recordHour < 0):
                        print("Record hour is negative, use '0:0' to record at 24 o'clock, midnight, at the same day.")
                        continue

                    if (recordMinute > 59):
                        print("Record time in minutes can only be smaller than 60 minutes")
                        continue
                    elif (recordMinute < 0):
                        print("Record minutes are negative. Use a positive value like '20:15'")
                        continue


                    timeNow = time.time()
                    timestampRecord = time.strptime("%d.%d.%d %d:%d:00" %(recordDay, recordMonth, recordYear, recordHour, recordMinute), "%d.%m.%Y %H:%M:%S")


                    if (timeNow > time.mktime(timestampRecord)):
                        print("Time is in the past. Repeat input.")
                        continue

                    if input("Is the record time correct, Y/N ? ").lower() != 'y':
                        continue

                    RECORDING_TIME = time.mktime(timestampRecord)
                    break
                else:
                    print("Invald time input.")
                    continue

            while True:
                recordDuration = input('\nEnter duration in minutes to record. For example 90 to capture 1 1/2 hours, "x" to cancel\n')
                if recordDuration.lower() == 'x':
                    print("Cancelled")
                    exit(1)
                elif int(recordDuration) < 0:
                    print("Negative recording duration, enter 30 for 30 minutes.")
                    continue

                print("Record duration: %d minutes " %(int(recordDuration)))

                if input("Is the record time correct, Y/N ? ").lower() != 'y':
                    continue

                RECORDING_DURATION = int(recordDuration)
                break


            while True:
                RECORDING_NAME = input('\nEnter name for the recording file. Enter "x" to cancel\n')
                if recordDuration.lower() == 'x':
                    print("Cancelled")
                    exit(1)

                RECORDING_NAME = RECORDING_NAME.replace(';', '-')
                print("Record name: %s" % (RECORDING_NAME))
                if input("Is the record name correct, Y/N ? ").lower() != 'y':
                    continue

                break



            print("\nWhat will be done?\nRecording channel \"%s\" @ %s." %(RECORDING_CHANNEL_NAME, RECORDING_CHANNEL_URL))
            print("On %s at %s o'clock for %d minutes.\nRecording to file: \"%s\"" %(time.strftime("%d.%m.%Y", time.localtime(RECORDING_TIME)), time.strftime("%H:%M:%S", time.localtime(RECORDING_TIME)), RECORDING_DURATION, RECORDING_NAME))

            confirm = input("\nAre all information correct, Y/N? ")
            if confirm.lower() != 'y' or confirm == '':
                continue

            scheduleFileName = "%d.rec" %(time.time())
            with open(scheduleFileName, 'w', encoding=PLAYLIST_ENCODING) as scheduleFile:
                scheduleFile.write("%d;%d;%s;%s;%s;%s\n" % (RECORDING_TIME, RECORDING_DURATION, RECORDING_CHANNEL_NAME, RECORDING_NAME, PLAYLIST_URL, RECORDING_CHANNEL_URL))

                if USE_LINUX_AT:
                    try:
                        if not ISWEBDL:
                            scheduleCommand = 'echo "python3 %s/rec.py -ldl%s -pf%s" | at -M %s' % (APP_FOLDER, scheduleFileName, PLAYLIST_FILE, time.strftime("%H:%M %d.%m.%Y", time.localtime(RECORDING_TIME)))
                        else:
                            scheduleCommand = 'echo "python3 %s/rec.py -dl%s" | at -M %s' % (APP_FOLDER, scheduleFileName, time.strftime("%H:%M %d.%m.%Y", time.localtime(RECORDING_TIME)))
                        subprocess.Popen(scheduleCommand, shell=True)
                        print("Scheduled task.\nBye bye.")
                        exit(1)
                    except Exception as error:
                        print('Error scheduling task using "at".')
                        raise error

                if USE_WINDOWS_TASK_SCHEDULER:
                    taskFileName = 'task_%s_%s.xml' %(RECORDING_NAME, RECORDING_TIME)
                    with open(taskFileName, 'w') as taskFile:
                        try:
                            TASK_XML = TASK_XML.replace('_REGISTRATION_DATE_', getXmlTimeString(time.time()))
                            TASK_XML = TASK_XML.replace('_TIMESTART_', getXmlTimeString(RECORDING_TIME))
                            TASK_XML = TASK_XML.replace('_TIMEEND_', getXmlTimeString(RECORDING_TIME + (24 * 60 * 60)))
                            TASK_XML = TASK_XML.replace('_COMMAND_', 'python %s\\rec.py' %(APP_FOLDER))
                            if not ISWEBDL:
                                TASK_XML = TASK_XML.replace('_COMMANDARGS_', '-ldl%s -pf%s' %(scheduleFileName, PLAYLIST_FILE))
                            else:
                                TASK_XML = TASK_XML.replace('_COMMANDARGS_', '-dl%s' %(scheduleFileName))

                            taskFile.write(TASK_XML)

                            scheduleCommand = 'schtasks /Create /TN ' + taskFileName +' /XML ' + taskFileName + ' /F'
                            subprocess.Popen(scheduleCommand, shell=True)
                            print("Scheduled task.\nBye bye.")
                            exit(1)
                        except Exception as error:
                            print('Error scheduling task with Task Schedular.')
                            raise error

                break

            print('Error writing schedule file for recording job.')

    else:
        hasPlaylist = False
        scheduleFile = None

        for opt in sys.argv:
            if opt.startswith('-dl'):
                scheduleFileName = opt.replace('-ldl', '', 1).replace('-dl', '', 1)

                if os.path.exists(scheduleFileName):
                    with open(scheduleFileName, "r", encoding=PLAYLIST_ENCODING, errors='strict') as scheduleFile:
                        scheduleData = scheduleFile.read().rstrip().split(';', 6)
                        PLAYLIST_URL = scheduleData[4]

                if downloadPlaylist(PLAYLIST_URL):
                    hasPlaylist = True
                    print("Playlist succesfully downloaded.\n")
                else:
                    print("Error downloading playlist from \"%s\".\nExiting." %(PLAYLIST_URL))
                    exit(1)

        for opt in sys.argv:
            if opt.startswith('-dl'):
                with open(PLAYLIST_FILE, "r", encoding=PLAYLIST_ENCODING, errors='strict') as playlistFile:
                    data = playlistFile.readlines()

                    currentChannel = None
                    for line in data:
                        line = bytes(line, PLAYLIST_ENCODING, 'strict').decode(PLAYLIST_ENCODING, 'strict').rstrip()
                        if line.startswith('#EXTINF:'):
                            try:
                                currentChannel = list(re.findall(r'tvg-name\=\"(.[^"]*)', line))[0].strip()
                            except:
                                currentChannel = line.split(',',2)[1].strip()

                            if line.find('radio="true"') != -1 or line.lower().find('webradio') != -1:
                                currentChannel += ' (Radio)'

                        elif (currentChannel != None and line.startswith('http')):
                            if currentChannel.startswith('*'):
                                continue
                            PLAYLIST_CHANNELS[currentChannel] = bytes(line, PLAYLIST_ENCODING).decode(PLAYLIST_ENCODING, 'strict').rstrip()
                            currentChannel = None

                    scheduleFileName = opt.replace('-dl', '', 1)
            elif opt.startswith('-ldl'):
                hasPlaylist = False
                for subopt in sys.argv:
                    if subopt.startswith('-pf'):
                        PLAYLIST_FILE = subopt.replace('-pf', '', 1)
                        hasPlaylist = True
                        break

                with open(PLAYLIST_FILE, "r", encoding=PLAYLIST_ENCODING, errors='strict') as playlistFile:
                    data = playlistFile.readlines()

                    currentChannel = None
                    for line in data:
                        if line.startswith('#EXTINF:'):
                            try:
                                currentChannel = list(re.findall(r'tvg-name\=\"(.[^"]*)', line))[0]
                            except:
                                currentChannel = line.split(',',2)[1].rstrip()

                            currentChannel = bytes(currentChannel, PLAYLIST_ENCODING).decode(PLAYLIST_ENCODING, 'strict').rstrip()

                            if line.find('radio="true"') != -1 or line.lower().find('webradio') != -1:
                                currentChannel += ' (Radio)'

                        elif (currentChannel != None and line.startswith('http')):
                            if currentChannel.startswith('*'):
                                continue
                            PLAYLIST_CHANNELS[currentChannel] = bytes(line, PLAYLIST_ENCODING).decode(PLAYLIST_ENCODING, 'strict').rstrip()
                            currentChannel = None

                sortedChannels = sorted(PLAYLIST_CHANNELS.keys())

        scheduleFile = None
        if os.path.exists(scheduleFileName):
            scheduleData = None

            with open(scheduleFileName, "r", encoding=PLAYLIST_ENCODING, errors='strict') as scheduleFile:
                scheduleData = scheduleFile.read().rstrip().split(';', 6)

            if scheduleData != None:
                try:
                    try:
                        channelURL = PLAYLIST_CHANNELS[currentChannel]
                    except:
                        channelURL = scheduleData[-1]

                    if scheduleData[2].find(' (Radio)') != -1:
                        commandString = FFMPEG_DIRECTORY + " -user_agent \"%s\" -i '%s' -t %d:%d:00 '%s.mp3'" %(USER_AGENT, channelURL, int(scheduleData[1]) / 60, int(scheduleData[1]) % 60, scheduleData[3].rstrip())
                    else:
                        commandString = FFMPEG_DIRECTORY + " -user_agent \"%s\" -i '%s' -t %d:%d:00 -vc:copy -o '%s.mkv'" % (USER_AGENT, channelURL, int(scheduleData[1]) / 60, int(scheduleData[1]) % 60, scheduleData[3].rstrip())
                        print(commandString)
                    subprocess.Popen(commandString, shell=True)
                except ValueError as error:
                    print('No channel information found.')
                    raise error
